/*
    ext2online.c -- resizes a mounted ext2 fs
    Copyright (C) 1999 Andreas Dilger <adilger@enel.ucalgary.ca>
    Copyright (C) 1999 Lennert Buytenhek <lbuijten@cs.leidenuniv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  ext2-online resize patch for the linux kernel 2.0, 2.2, or 2.3.
*/

static const char _ext2online_c[] = "$Id: ext2online.c,v 1.11 2000/03/08 10:35:07 buytenh Exp $";

#include "config.h"

#ifndef HAVE_LINUX_EXT2_FS_H
#error This program only works on linux
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <mntent.h>
#include <sys/mount.h> /* mount.h may define these next two, but if not... */
#if !defined(MS_MGC_VAL) || !defined(MS_REMOUNT)
#include <linux/fs.h>
#endif
#include <signal.h>
#include "ext2.h"

#ifndef MNTTYPE_EXT2
#define MNTTYPE_EXT2 "ext2"
#endif

char			*progname;
int			 opt_debug = 0;
int			 opt_verbose = 0;
int			 opt_ignore = 0;

static void print_version(FILE *outfile)
{
	fprintf(outfile, "ext2online v%s - 1999/10/14 for EXT2FS %s\n",
		VERSION, EXT2FS_VERSION);
}

static void usage(FILE *outfile)
{
	fprintf(outfile,
		"usage: %s [-d] [-i] [-v] device [new_size[bkmgt]]\n"
		"\t-d, --debug    : turn debug info on\n"
		"\t-i, --ignore   : ignore device size check\n"
		"\t-v, --verbose  : be verbose\n"
		"\tnew_size is in ext2 filesystem blocks (1k, 2k, or 4k),\n"
		"\tdisk Blocks (512 byte), Kilo-, Mega-, Giga-, or Terabytes\n",
		progname);
	exit(1);
}

static void parse_args(int *argc, char **argv[])
{
	static struct option long_options[] = {
		{"debug", 0, 0, 'd'},
		{"ignore", 0, 0, 'i'},
		{"verbose", 0, 0, 'v'},
		{NULL, 0, 0, 0}};
	int c;
	int dummy;

	while ((c = getopt_long(*argc, *argv, "div", long_options, &dummy)) != -1)
	{
		switch (c)
		{
		case 'd': opt_debug = 1; break;
		case 'i': opt_ignore = 1; break;
		case 'v': opt_verbose = 1; break;
		default:  fprintf(stderr, "%s: unknown option %c\n",
				  progname, c);
			  usage(stderr);
		}
	}

	*argc -= optind;
	*argv += optind;
}

/* Find the mountpoint or device from the string in *dev */
static void find_mount(char *filename, char **dev, char **dir, char **type,
		       char **opts)
{
	FILE		*tabfile;
	struct mntent	*ent;

	tabfile = setmntent(filename, "r");

	while  ((ent = getmntent(tabfile)) != NULL)
	{
		if (strcmp(*dev, ent->mnt_fsname) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = (char *)malloc(strlen(ent->mnt_dir) + 1);
			if (*dir == NULL)
				*dir = ent->mnt_dir;
			else
				strcpy(*dir, ent->mnt_dir);
			break;
		}
		else if (strcmp(*dev, ent->mnt_dir) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = *dev;
			*dev = (char *)malloc(strlen(ent->mnt_fsname) + 1);
			if (*dev == NULL)
				*dev = ent->mnt_dir;
			else
				strcpy(*dev, ent->mnt_fsname);
			break;
		}
	}
	if (*dir != NULL && ent != NULL)
	{
		*type = (char *)malloc(strlen(ent->mnt_type) + 1);
		if (*type == NULL)
			*type = ent->mnt_type;
		else
			strcpy(*type, ent->mnt_type);
		/* Leave room at end for ",resize=NNNNNNNNNN:NNNN" parameter */
		*opts = (char *)malloc(strlen(ent->mnt_opts) + 25);
		if (*opts == NULL)
		{
			fprintf(stderr,
			"%s: no memory for %s mount options, resize aborted\n",
				progname, *dev);
			*dir = NULL;
		}
		else
		{
			char *index, *index2;

			if (hasmntopt(ent, MNTOPT_RO))
			{
				fprintf(stderr,
				"%s: %s is mounted read-only, resize aborted\n",
					progname, *dev);
				*dir = NULL;
			}

			/* Skip the "rw" and "nosuid" options, keep others */
			if ((index = hasmntopt(ent, MNTOPT_NOSUID)) != NULL)
			{
				if ((index2 = strchr(index, ',')) == NULL)
					index[0] = '\0';
				else
					while ((*index++ = *++index2) != '\0');
			}

			if ((index = hasmntopt(ent, MNTOPT_RW)) != NULL)
			{
				if ((index2 = strchr(index, ',')) == NULL)
					*index = '\0';
				else
					while ((*index++ = *++index2) != '\0');
			}

			if (ent->mnt_opts[0] == '\0')
				*opts[0] = '\0';
			else
				sprintf(*opts, "%s,", ent->mnt_opts);
		}
	}
	endmntent(tabfile);
}


/* Check that all of the needed group descriptor blocks are reserved */
static int ext2_check_newgdblocks(struct ext2_fs *fs, struct ext2_inode *inode,
				  blk_t newgdblocks)
{
	int	group;
	blk_t	block;
	blk_t	gdb;

	for (group = 0, gdb = fs->sb.s_first_data_block + 1;
	     group < fs->numgroups; group++, gdb += fs->sb.s_blocks_per_group)
	for (block = fs->gdblocks; block < newgdblocks; block++) {
		if (ext2_is_group_sparse(fs, group)) {
			if (fs->opt_debug)
				printf("checking for group block %d in Bond\n",
				       gdb);

			if (ext2_do_inode(fs, inode, gdb + block,
					  EXT2_ACTION_FIND) < 0) {
				fprintf(stderr,
					"%s: group %d, block %u not reserved\n",
					progname, group, gdb + block);
				return 0;
			}
		}
	}
	return 1;
}

#define mark_block(bh, offset) (bh)->data[(offset)>>3] |= _bitmap[(offset)&7]

/* Create group descriptors, bitmaps, inodes for each new group */
static blk_t ext2_make_group(struct ext2_fs *fs, struct ext2_inode *inode,
			     int group, blk_t blocks, blk_t newgdblocks,
			     blk_t stride, blk_t resgdblocks)
{
	struct ext2_buffer_head	*bh;
	struct ext2_group_desc	*gdp;
	blk_t			 start;
	blk_t			 bboffset, iboffset, itoffset;
	int			 max_inodes;
	int			 i;

	if (fs->opt_verbose)
	{
		printf("%s: creating group %3d with %5u blocks "
		       "(rsvd = %2u, newgd = %2u)\r",
		       progname, group, blocks, resgdblocks, newgdblocks);
		fflush(stdout);
	}

	if (blocks < fs->adminblocks + 50)
	{
		/* This would only happen on the last group */
		fprintf(stderr, "\n%s: unable to use %u blocks "
			"at end of device\n", progname, blocks);
		return 0;
	}

	start = fs->sb.s_first_data_block + group * fs->sb.s_blocks_per_group;

	itoffset = fs->itoffset;
	if (fs->opt_debug)
		printf("\nusing itoffset of %d\n", fs->itoffset);
	if (itoffset + fs->inodeblocks > blocks)
	{
		printf("\n%s: resetting metadata offset for group %d\n",
			progname, group);
		itoffset = 3 + ext2_is_group_sparse(fs, group) ?
			   newgdblocks + resgdblocks : 0;
	}

	/* Create the primary copy of the group descriptor table */
	bh = ext2_bread(fs, fs->sb.s_first_data_block +
			howmany((group + 1) * sizeof(struct ext2_group_desc),
				fs->blocksize));
	gdp = (struct ext2_group_desc *)bh->data +
	      (group % sizeof(struct ext2_group_desc));

	if (stride)
	{
		blk_t next = itoffset + fs->inodeblocks;
		blk_t newoff = next + stride * group;

		if (fs->opt_debug)
			printf("stride %d, next %d, newoff %d\n",
			       stride, next, newoff);

		stride = newoff >= blocks ? next : newoff;
	}

	if (ext2_is_group_sparse(fs, group))
	{
		bboffset = stride ? stride : itoffset - 2;
		gdp->bg_free_blocks_count = 0;
		if (fs->opt_debug)
			printf("normal bboffset %d\n", bboffset);
	}
	else
	{
		bboffset = stride ? stride : 0;
		gdp->bg_free_blocks_count = 1 + newgdblocks + resgdblocks;
		if (fs->opt_debug)
			printf("sparse bboffset %d (free %d)\n", bboffset,
			       gdp->bg_free_blocks_count);
	}

	iboffset = bboffset + 1;

	gdp->bg_block_bitmap = start + bboffset;
	gdp->bg_inode_bitmap = start + iboffset;
	gdp->bg_inode_table  = start + itoffset;
	gdp->bg_free_blocks_count += blocks - fs->adminblocks;
	gdp->bg_free_inodes_count = fs->sb.s_inodes_per_group;
	gdp->bg_used_dirs_count = 0;
	gdp->bg_pad = 0;
	gdp->bg_reserved[0] = 0;
	gdp->bg_reserved[1] = 0;
	gdp->bg_reserved[2] = 0;

	/* Write only this specific group to disk, so we don't corrupt
	 * other group descriptors within this block.
	 */
	fs->devhandle->ops->direct_write(fs->devhandle->cookie, gdp,
		(fs->sb.s_first_data_block + 1) * fs->blocksize +
		group * sizeof(struct ext2_group_desc),
		sizeof(struct ext2_group_desc));
	/* Update the in-memory copy of the group descriptor as well */
	memcpy(&fs->gd[group], gdp, sizeof(struct ext2_group_desc));
	ext2_brelse(bh, 0);

	/* Clear metadata blocks to show all blocks/inodes unused */
	if (fs->opt_debug)
	{
		printf("clear/mark block bitmap 0x%04x used\n",
			start + bboffset);
		printf("clear/mark inode bitmap 0x%04x used\n",
			start + iboffset);
		printf("clear/mark inode blocks 0x%04x-0x%04x used\n",
			start + itoffset,
			start + itoffset + fs->inodeblocks - 1);
	}

	ext2_zero_blocks(fs, start + bboffset, 1);
	ext2_zero_blocks(fs, start + iboffset, 1);
	ext2_zero_blocks(fs, start + itoffset, fs->inodeblocks);

	/* Mark in-use metadata blocks */
	bh = ext2_bread(fs, start + bboffset);
	bh->dirty = 1;
	mark_block(bh, bboffset);
	mark_block(bh, iboffset);
	for (i = itoffset; i < itoffset + fs->inodeblocks; i++)
		mark_block(bh, i);

	/* Mark metadata blocks used in block bitmap */
	if (ext2_is_group_sparse(fs, group))
	{
		if (fs->opt_debug)
		{
			printf("mark superblock 0x%04x used\n", start);
			printf("mark group desc. 0x%04x-0x%04x used\n",
				start + 1, start + newgdblocks);
			if (resgdblocks)
				printf("mark reserved descriptors "
					"0x%04x-0x%04x used\n",
					start + newgdblocks + 1,
					start + newgdblocks + resgdblocks);
		}

		mark_block(bh, 0);		/* superblock */
		for (i = 1; i <= newgdblocks; i++)
			mark_block(bh, i);	/* group descriptor blocks */
		for (; i <= newgdblocks + resgdblocks; i++)
		{
			mark_block(bh, i);	/* reserved GDT blocks */
			ext2_do_inode(fs, inode, start + i, EXT2_ACTION_ADD);
		}
	}

	/* Mark unavailable blocks at end of group used */
	if (fs->opt_debug && blocks != fs->blocksize * 8)
		printf("mark end blocks 0x%04x-0x%04x used\n",
			start + blocks, start + fs->blocksize * 8 - 1);

	for (i = blocks; i < fs->blocksize * 8; i++)
		mark_block(bh, i);
	ext2_brelse(bh, 0);

	/* Mark unavailable inodes at end of table used */
	max_inodes = fs->blocksize * 8;
	if (max_inodes != fs->sb.s_inodes_per_group)
	{
		if (fs->opt_debug)
			printf("mark %d unavailable end inodes used\n\n",
			       max_inodes - fs->sb.s_inodes_per_group);

		bh = ext2_bread(fs, iboffset + start);
		bh->dirty = 1;
		for (i = fs->sb.s_inodes_per_group; i < max_inodes; i++)
			bh->data[i>>3] |= _bitmap[i&7];
		ext2_brelse(bh, 0);
	}

	return blocks;
}

/* Create new groups if necessary to handle new filesystem size */
static blk_t ext2_online_primary(struct ext2_fs *fs, struct ext2_inode *inode,
				 blk_t newsize, int newgroups,
				 blk_t newgdblocks, blk_t resgdblocks)
{
	int			 group;
	int			 stride = 0;
	blk_t			 realsize;

	if (newsize <= fs->sb.s_blocks_count)
		return newsize;

	if (fs->opt_debug)
	{
		printf("\n%d old groups, %d blocks\n", fs->numgroups,
			fs->gdblocks);
		printf("%d new groups, %d blocks\n", newgroups,
			newgdblocks);
	}

	/* We need to use more GDT blocks - see if they are available */
	if (newgdblocks > fs->gdblocks &&
	    !ext2_check_newgdblocks(fs, inode, newgdblocks))
			return 0;

	/* Try to keep RAID spacings.  We don't check the first few
	 * groups as ext2prepare might have moved the block and inode
	 * bitmaps.
	 */
	if (fs->numgroups > 2)
	{
		int last = fs->numgroups - 1;
		int bpg = fs->sb.s_blocks_per_group;

		stride = fs->gd[last].bg_block_bitmap -
			 fs->gd[last - 1].bg_block_bitmap - bpg;
		if (stride != fs->gd[last - 1].bg_block_bitmap -
			      fs->gd[last - 2].bg_block_bitmap - bpg) {
			if (fs->opt_debug)
				printf("This is an odd filesystem, the block "
				       "bitmaps are located differently.\n");
			stride = 0;
		}
		if (fs->opt_debug)
			printf("Using a RAID stride value of %d\n", stride);
	}

	/* Fill last existing group */
	realsize = min(fs->numgroups * fs->sb.s_blocks_per_group +
			fs->sb.s_first_data_block, newsize);
	if (fs->opt_debug && realsize > fs->sb.s_blocks_count)
		printf("Filling last group to %u blocks\n", realsize);

	fs->adminblocks += newgdblocks + resgdblocks - fs->gdblocks;
	/* Add new groups */
	for (group = fs->numgroups; group < newgroups; group++)
	{
		blk_t blocks;

		blocks = min(fs->sb.s_blocks_per_group, newsize - realsize);
		realsize += ext2_make_group(fs, inode, group, blocks,
					    newgdblocks, stride, resgdblocks);
	}

	return realsize;
}

/* Copy primary superblock, group descriptors to all "sparse" groups after
 * we are sure that the resize has worked
 */
static void ext2_online_finish(struct ext2_fs *fs, struct ext2_inode *inode,
			       blk_t newblocks, int newgroups,
			       blk_t newgdblocks, blk_t resgdblocks)
{
	int			 group;
	blk_t			 block;
	blk_t			 gdb;

	/* Remove blocks as required from Bond inode */
	for (group = 0, gdb = fs->sb.s_first_data_block + 1;
	     group < fs->numgroups;
	     group++, gdb += fs->sb.s_blocks_per_group)
		if (ext2_is_group_sparse(fs, group))
			for (block = fs->gdblocks; block < newgdblocks; block++)
			{
				if (fs->opt_debug)
					printf("Checking for group block %d "
					       "in Bond\n", gdb);
				ext2_do_inode(fs, inode, gdb + block,
					      EXT2_ACTION_DELETE);
			}

	/* Save the new number of reserved GDT blocks in the resize inode */
	inode->i_gid = resgdblocks;

	/* Re-read kernel-updated superblock from disk, and fix up the rest
	 * of the fs structure so that the backup superblock copies are OK.
	 */
	fs->devhandle->ops->direct_read(fs->devhandle->cookie, &fs->sb,
				 1024, 1024);
	fs->gdblocks = newgdblocks;
	fs->numgroups = newgroups;
	fs->metadirty = EXT2_META_BACKUP;
}

static blk_t get_reserved(struct ext2_fs *fs, struct ext2_inode *inode,
			  blk_t newgdblocks)
{
	int	resgdblocks;

	/* How many blocks to reserve for each group future GDT expansion.
	 * This assumes Bond has the right number of blocks assigned.  It
	 * may not be reliable unless e2fsck starts fixing Bond for us.
	 */
	if (inode->i_blocks == 0)
		return 0;

	/* This is the new/preferred method of storing reserved blocks */
	if (inode->i_gid)
		resgdblocks = inode->i_gid + fs->gdblocks - newgdblocks;

	/* We want to also allow the old version in case someone prepared a
	 * filesystem and they now want to resize with the new code.  This
	 * can be removed in a future release.
	 */
	/* FIXME: This doesn't count any additional indirect metadata blocks
	 *        (rounding down should cover this in most cases), and doesn't
	 *        work for sparse filesystems.
	 */
	else if (!fs->sparse)
		resgdblocks = ((inode->i_blocks >> (fs->logsize - 9)) -
			       (inode->i_block[EXT2_IND_BLOCK] ? 1 : 0) -
			       (inode->i_block[EXT2_DIND_BLOCK] ? 1 : 0)) /
			      fs->numgroups + fs->gdblocks - newgdblocks;
	else
	/* FIXME: this only works for a normal filesystem that has been
	 * modified by ext2prepare only.
	 */
		resgdblocks = fs->itoffset - newgdblocks - 3;

	if (resgdblocks > 200 || resgdblocks < 0) { /* Bogosity test */
		fprintf(stderr,
			"%s: wrong resgdblocks calculated (%d), using 0.\n"
			"\tyou may need ext2prepare before using %s again\n",
			progname, resgdblocks, progname);
		resgdblocks = 0;
	}

	if (fs->opt_debug)
		printf("using %u reserved GDT blocks\n", resgdblocks);

	return resgdblocks;
}

int main(int argc, char *argv[])
{
	char			*dev, *type, *opts, *dir = NULL;
	struct ext2_dev_handle	*handle;
	struct ext2_fs		*fs;
	struct ext2_inode	 inode;
	__u32			 i_mtime;
	int			 newgroups;
	blk_t			 newsize = 0, devsize = 0, realsize = 0;
	blk_t			 newgdblocks, resgdblocks;
	char			 mod = '\0';

	progname = strrchr(argv[0], '/') ? strrchr(argv[0], '/') + 1 : argv[0];
	print_version(stdout);
	parse_args(&argc, &argv);

	if (argc < 1 || argc > 2)
		usage(stderr);

	dev = argv[0];

	find_mount(MOUNTED, &dev, &dir, &type, &opts);
	if (dir == NULL)
	{
		fprintf(stderr, "%s: can't find %s, is it mounted?\n",
			progname, dev);
		usage(stderr);
	}

	handle = ext2_make_dev_handle_from_file(dev);
	if (handle == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, progname);
		return 1;
	}

	/* See if the user has specified a filesystem size */
	if (argc == 2)
	{
		sscanf(argv[1], "%i%c", &newsize, &mod);

		switch (mod)	/* Order of these options is important!! */
		{
			case 't':
			case 'T': newsize *= 1024; /* no break */
			case 'g':
			case 'G': newsize *= 1024; /* no break */
			case 'm':
			case 'M': newsize *= 1024; /* no break */
			case 'k': /* Need to open fs to find block size */
			case 'K': break;
			case  0 : break;
			case 'b':
			case 'B': newsize /= 2; break;
			default: usage(stderr);
		}

		newgroups = howmany(newsize, 8192); /* worst case 1k blocks */

		if (ext2_max_groups < newgroups)
			ext2_max_groups = newgroups;
	}

	if ((fs = ext2_open(handle, EXT2_VALID_FS)) == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, progname);
		return 1;
	}
	fs->online = 1;
	fs->opt_debug = opt_debug;
	fs->opt_verbose = opt_verbose;

	/* Convert the filesystem size from KB to number of blocks */
	if (mod != '\0')
		newsize /= fs->blocksize / 1024;

	devsize = fs->devhandle->ops->get_size(fs->devhandle->cookie);

	if (argc == 1)	/* Resize to the full size if none is given */
		newsize = devsize;
	else if (newsize > devsize && !opt_ignore)
	{
		/* We could put a system() call here to lvextend with the
		 * appropriate parameters to grow the LV before the resize.
		 */
		fprintf(stderr, "%s: %s has %u blocks not %u, resize aborted\n",
			progname, dev, devsize, newsize);
		return 1;
	}
	else if (newsize <= fs->sb.s_blocks_count)
	{
		/* Not currently, but it would be possible if the groups at
		 * the end of the FS are not in use at all (no inode, blocks)
		 * We could even do a system() call to lvreduce after the
		 * mount() call returned, when we can get a status code back...
		 */
		fprintf(stderr, "%s: %s has %u blocks cannot shrink to %u\n",
			progname, dev, fs->sb.s_blocks_count, newsize);
		return 1;
	}

	if (fs->opt_debug)
		printf("device size %d, specified %d\n", devsize, newsize);

	newgroups = howmany(newsize - fs->sb.s_first_data_block,
			    fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc),
			      fs->blocksize);
	/* Bring in Mr. Bond */
	ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);
	i_mtime = inode.i_mtime;

	/* How many blocks to reserve in each group for future GDT expansion */
	resgdblocks = get_reserved(fs, &inode, newgdblocks);

	if ((realsize = ext2_online_primary(fs, &inode, newsize, newgroups,
					    newgdblocks, resgdblocks)) == 0)
	{
		fprintf(stderr, "\n%s: unable to resize %s past %u blocks\n",
			progname, dev, fs->gdblocks * fs->sb.s_blocks_per_group*
			fs->blocksize / sizeof(struct ext2_group_desc));
		return 2;
	}
	else if (realsize == fs->sb.s_blocks_count)
	{
		fprintf(stderr, "%s: new size for %s is same as current (%u)\n",
			progname, dev, newsize);
		return 2;
	}

	printf("\n%s: resizing to %u blocks\n", progname, realsize);

	if (i_mtime != inode.i_mtime)
	{
		ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);
		i_mtime = inode.i_mtime;
	}
	ext2_sync(fs);

	/* Add the resize option to the end of the other mount options */
	sprintf(opts + strlen(opts), "resize=%u", realsize);
	if (resgdblocks)
		sprintf(opts + strlen(opts), ":%u", resgdblocks);

	if (fs->opt_debug)
	{
		printf("Calling mount() for %s(%s) with %s\n", dev, dir, opts);
#if 0
		printf("Press ENTER to continue:\n");
		mod = getchar();
#endif
	}

	/* Ignore basic interrupts at this point, so we have a chance to
	 * clean up.  Before now, if the resize was interrupted, it didn't
	 * matter, as we haven't made any real changes to the filesystem.
	 */
	{
		struct sigaction act;
		act.sa_handler = SIG_IGN;
		act.sa_flags = SA_RESTART;

		sigaction(SIGHUP, &act, NULL);
		sigaction(SIGINT, &act, NULL);
		sigaction(SIGQUIT, &act, NULL);
	}

	if (mount(dev, dir, type, MS_MGC_VAL | MS_REMOUNT, opts) < 0)
	{
		fprintf(stderr,"%s: resize failed while in kernel\n", progname);
		perror(progname);
		return 3;
	}

	/* Clean up the rest of the filesystem and the Bond inode */
	ext2_online_finish(fs, &inode, realsize, newgroups, newgdblocks,
			   resgdblocks);
	if (i_mtime != inode.i_mtime)
		ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);
	ext2_close(fs);

	return 0;
}
