/*
    ext2prepare.c -- prepares an ext2 fs for an online resize
    Copyright (C) 1999 Andreas Dilger <adilger@enel.ucalgary.ca>
    Copyright (C) 1999 Lennert Buytenhek <lbuijten@cs.leidenuniv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  online resize patch for the linux kernel.
*/

static const char _ext2prepare_c[] = "$Id: ext2prepare.c,v 1.9 2000/03/08 10:35:09 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include <sys/stat.h>
#include "ext2.h"

char	*prog;
blk_t	 newsize;
int	 newgroups;
blk_t	 newgdblocks;
int	 resblocks;
int	 u32perblock;
int	 opt_debug = 0;
int	 opt_verbose = 0;

#define EXT2_IND_UNCHANGED	1
#define EXT2_IND_CHANGED	2

void print_version(FILE *outfile)
{
	fprintf(outfile, "ext2prepare v%s - 1999/10/02 for EXT2FS %s\n",
		VERSION, EXT2FS_VERSION);
}

void usage(FILE *outfile)
{
	fprintf(outfile, "usage: %s device max_new_size[bkmgt]\n", prog);
	fprintf(outfile,
"       where new_size is in ext2 filesystem blocks (1k, 2k, or 4k),\n"
"       disk Blocks (512), Kilo-, Mega-, Giga-, or Terabytes\n");
	exit(1);
}

void parse_args(int *argc, char **argv[])
{
	static struct option long_options[] = {
		{"debug", 0, 0, 'd'},
		{"verbose", 0, 0, 'v'},
		{NULL, 0, 0, 0}};
	int c;
	int dummy;

	while ((c = getopt_long(*argc, *argv, "dfuv", long_options, &dummy)) != -1)
	{
		switch (c)
		{
		case 'd': opt_debug = 1; break;
		case 'v': opt_verbose = 1; break;
		default:  usage(stderr);
		}
	}

	*argc -= optind;
	*argv += optind;
}

/* Calculate the maximum number of blocks that we need to reserve in Bond */
blk_t calc_blocks(struct ext2_fs *fs)
{
	blk_t reserved;

	newgroups = howmany(newsize - fs->sb.s_first_data_block,
			 fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc),
			   fs->blocksize);

	/* For a non-sparse filesystem, the maximum number of blocks to
	 * reserve happens when we have half of the filesystem full.
	 * If we are past half full, then we need fewer reserved blocks.
	 */
	if (!fs->sparse)
		reserved = (newgroups - max(fs->numgroups, newgroups / 2)) *
			   (newgdblocks - max(fs->gdblocks, newgdblocks / 2));
	/* For a sparse filesystem, the number of blocks to reserve is not a
	 * linear function of the filesystem size.  Instead of actually working
	 * out the exact maximum, we have calculated that the maximum possible
	 * number of groups with GDT blocks in a sparse filesystem is 49, but
	 * it will typically be much less.
	 */
	else
		reserved = (newgdblocks - fs->gdblocks) * 49;

	if (fs->opt_debug)
		fprintf(stderr, "calc_blocks: "
			"newgroups = %d, newgdblocks = %d, reserved = %d\n",
			newgroups, newgdblocks, reserved);

	return reserved;
}


blk_t getfreezeroedblock(struct ext2_fs *fs)
{
	int i;
	blk_t offset;

	for (i = 0, offset = fs->sb.s_first_data_block; i < fs->numgroups;
	     i++, offset += fs->sb.s_blocks_per_group)
	{
		blk_t j;

		/* Don't allocate inode blocks in range where we may need
		 * to move GDT blocks later */
		for (j = fs->itoffset + fs->inodeblocks + newgdblocks;
		     j < fs->sb.s_blocks_per_group; j++)
		{
			if (!ext2_get_block_state(fs, offset + j))
			{
				struct ext2_buffer_head *bh;

				ext2_set_block_state(fs, offset + j, 1, 1);
				bh = ext2_bcreate(fs, offset + j);
				bh->dirty = 1;
				ext2_brelse(bh, 0);

				return offset + j;
			}
		}
	}

	return 0;
}

/* allocate an indirect block, if necessary */
int check_inode_ind(struct ext2_fs *fs, blk_t *block, blk_t *blocks, int decr)
{
	if (decr)
		resblocks -= u32perblock;

	if (*block == 0) {
		*block = getfreezeroedblock(fs);
		if (*block == 0)
			return 0;
		*blocks += 2;
		if (fs->opt_debug) {
			fprintf(stderr, "check_inode_ind: add indirect 0x%04x ",
				*block);
			if (decr)
				fprintf(stderr, "(%d remain)",
					(resblocks + u32perblock - 1) /
					u32perblock);
			else
				fprintf(stderr, "(double indirect)");
			fprintf(stderr, "\n");
		}
		return EXT2_IND_CHANGED;
	}

	return EXT2_IND_UNCHANGED;
}

int check_inode_dind(struct ext2_fs *fs, blk_t *block, blk_t *blocks)
{
	struct ext2_buffer_head	*bh;
	__u32			*udata;
	int			 change = 0;
	int			 res;
	int			 i;

	if ((res = check_inode_ind(fs, block, blocks, 0)) == 0)
		return 0;
	change |= res;
	bh = ext2_bread(fs, *block);
	udata = (__u32 *)bh->data;
	for (i = 0; i < u32perblock && resblocks > 0; i++, udata++) {
		if ((res = check_inode_ind(fs, udata, blocks, 1)) == 0)
			return 0;
		change |= res;
	}
	bh->dirty = change & EXT2_IND_CHANGED ? 1: 0 ;
	ext2_brelse(bh, 0);

	return change;
}

int make_inode_ind(struct ext2_fs *fs, struct ext2_inode *inode)
{
	resblocks -= EXT2_NDIR_BLOCKS;
	if (resblocks > 0 &&
	    check_inode_ind(fs, inode->i_block + EXT2_IND_BLOCK,
			    &inode->i_blocks, 1) == 0) {
		fprintf(stderr, "%s: can't reserve indirect block\n", prog);
		return 0;
	}

	if (resblocks > 0 &&
	    check_inode_dind(fs, inode->i_block + EXT2_IND_BLOCK,
			    &inode->i_blocks) == 0) {
		fprintf(stderr, "%s: can't reserve dindirect blocks\n", prog);
		return 0;
	}

	if (resblocks > 0) {
		fprintf(stderr, "%s: too many reserved blocks\n", prog);
		return 0;
	}
	return 1;
}

int create_resize_inode(struct ext2_fs *fs)
{
	int			diff;
	int			i;
	struct ext2_inode	inode;
	blk_t			block;

	ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);
	if (inode.i_mode == 0 || inode.i_links_count == 0)
	{
		if (fs->opt_verbose)
			printf("creating resize inode (%03d)\n",
				EXT2_RESIZE_INO);
		inode.i_mode = S_IFREG | S_IRUSR | S_IWUSR;
		inode.i_size = 0;
		inode.i_uid = inode.i_gid = 0;
		inode.i_atime = inode.i_ctime = time(NULL);
		inode.i_dtime = 0;
		inode.i_links_count = 1;
		inode.i_blocks = 0;
		inode.i_flags = EXT2_NODUMP_FL;
		inode.osd1.linux1.l_i_reserved1 = 0;
		for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
			inode.i_block[i] = 0;
		inode.i_block[EXT2_IND_BLOCK]  = 0;
		inode.i_block[EXT2_DIND_BLOCK] = 0;
		inode.i_block[EXT2_TIND_BLOCK] = 0;
		inode.i_version = 0;
		inode.i_file_acl = 0;
		inode.i_dir_acl = 0;
		inode.i_faddr = 0;
		inode.osd2.linux2.l_i_frag = 0;
		inode.osd2.linux2.l_i_fsize = 0;
		inode.osd2.linux2.i_pad1 = 0;
		inode.osd2.linux2.l_i_reserved2[0] = 0;
		inode.osd2.linux2.l_i_reserved2[1] = 0;
	}

	if (!make_inode_ind(fs, &inode))
		return 0;

	diff = newgdblocks - fs->gdblocks;
	/* we store the number of reserved blocks in the inode GID field */
	inode.i_gid = diff;
	for (i = 0, block = fs->sb.s_first_data_block + fs->gdblocks + 1;
	     i < fs->numgroups;
	     i++, block += fs->sb.s_blocks_per_group) {
		int j;

		if (!ext2_is_group_sparse(fs, i))
			continue;

		for (j = 0; j < diff; j++) {
			    if (ext2_do_inode(fs, &inode, block + j,
					      EXT2_ACTION_FIND) < 0 &&
				ext2_do_inode(fs, &inode, block + j,
					      EXT2_ACTION_ADD) < 0)
				return 0;
		}
	}

	inode.i_mtime = time(NULL);
	ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);

	return 1;
}


int ext2_prepare_fs(struct ext2_fs *fs)
{
	int status;

	if (fs->opt_verbose)
		printf("%s: prepare for %u blocks\n", prog, newsize);

	if (newsize <= fs->sb.s_blocks_count)
		return 0;

	if ((fs->relocator_pool =
	     (unsigned char *)malloc(ext2_relocator_pool_size << 10)) == NULL) {
		fprintf(stderr, "%s: error allocating relocator pool!\n", prog);
		return 0;
	}
	fs->relocator_pool_end = fs->relocator_pool +
				 (ext2_relocator_pool_size << 10);

	status = ext2_block_relocate(fs, newsize);

	if (status)
		status = ext2_metadata_push(fs, newsize);

	if (status)
		status = create_resize_inode(fs);

	free(fs->relocator_pool);
	fs->relocator_pool = NULL;
	fs->relocator_pool_end = NULL;

	return status;
}

int main(int argc, char *argv[])
{
	struct ext2_dev_handle	*handle;
	struct ext2_fs		*fs;
	blk_t			 maxres;
	char			 mod = '\0';
	char			*dev;

	prog = argv[0];
	print_version(stdout);
	parse_args(&argc, &argv);

	if (argc != 2)
		usage(stderr);

	dev = argv[0];
	handle = ext2_make_dev_handle_from_file(dev);
	if (handle == NULL) {
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	sscanf(argv[1], "%i%c", &newsize, &mod);

	switch (mod) {	/* Order of these options is important!! */
		case 't':
		case 'T': newsize *= 1024; /* no break */
		case 'g':
		case 'G': newsize *= 1024; /* no break */
		case 'm':
		case 'M': newsize *= 1024; /* no break */
		case 'k':
		case 'K': /* Need to open fs to find block size */
		case  0 : break;
		case 'b':
		case 'B': newsize /= 2; break;
		default: usage(stderr);
	}

	newgroups = howmany(newsize, 8192); /* worst case 1k blocks */

	if (ext2_max_groups < newgroups)
		ext2_max_groups = newgroups;

	if ((fs = ext2_open(handle, 0)) == NULL) {
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	fs->opt_debug = opt_debug;
	fs->opt_verbose = opt_verbose;

	if (mod != '\0')
		newsize /= fs->blocksize / 1024;

	/* For non-sparse 1k filesystems over 128GB, over one-quarter of
	 * the filesystem would be used by group descriptor blocks!
	 */
	if (fs->blocksize == 1024 && newsize > 1<<27 && !fs->sparse) {
		fprintf(stderr,
			"%s: won't resize 1k block filesystem over 128GB.\n"
			"\tToo much space would be wasted.\n", prog);
		return 1;
	} else if ((fs->blocksize == 4096 && newsize > 1<<30) ||
		   newsize > 1<<31) {
		fprintf(stderr,
			"%s: can't resize filesystem over 2TB.\n", prog);
		return 1;
	}

	u32perblock = fs->blocksize >> 2;
	resblocks = calc_blocks(fs);

	/* Maximum number of blocks we can reserve depends on block size,
	 * and the fact we do not use triple-indirect blocks.
	 */
	maxres = (1 + u32perblock) * u32perblock + EXT2_NDIR_BLOCKS;
	if (resblocks > maxres) {
		/* Round down to the next multiple of the group size */
		newsize &= ~(fs->sb.s_blocks_per_group - 1);
		resblocks = calc_blocks(fs);
		while (resblocks > maxres) {
			newsize -= fs->sb.s_blocks_per_group;
			resblocks = calc_blocks(fs);
		}
		fprintf(stderr,
		       "%s: can't prepare for filesystem over %d blocks\n"
		       "\tuntil the current filesystem grows larger.\n",
		       prog, newsize);
		return 1;
	}

	if (!ext2_prepare_fs(fs))
		return 2;

	printf("%s: prepared for maximum filesystem size of %d blocks\n", prog,
	       newgdblocks * (fs->blocksize / sizeof(struct ext2_group_desc)) *
	       fs->sb.s_blocks_per_group);
	ext2_close(fs);

	return 0;
}
